"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const glob = require("glob");
const path = require("path");
const index_1 = require("../index");
class Actions extends index_1.Initializer {
    constructor() {
        super();
        this.name = "actions";
        this.loadPriority = 410;
    }
    async initialize() {
        index_1.api.actions = {
            actions: {},
            versions: {},
            middleware: {},
            globalMiddleware: []
        };
        /**
         * Add a middleware component avaialable to pre or post-process actions.
         */
        index_1.api.actions.addMiddleware = (data) => {
            if (!data.name) {
                throw new Error("middleware.name is required");
            }
            if (!data.priority) {
                data.priority = index_1.api.config.general.defaultMiddlewarePriority;
            }
            data.priority = Number(data.priority);
            index_1.api.actions.middleware[data.name] = data;
            if (data.global === true) {
                index_1.api.actions.globalMiddleware.push(data.name);
                index_1.api.utils.sortGlobalMiddleware(index_1.api.actions.globalMiddleware, index_1.api.actions.middleware);
            }
        };
        index_1.api.actions.loadFile = async (fullFilePath, reload) => {
            if (reload === null) {
                reload = false;
            }
            const loadMessage = action => {
                if (reload) {
                    index_1.api.log(`action reloaded: ${action.name} @ v${action.version}, ${fullFilePath}`, "info");
                }
                else {
                    index_1.api.log(`action loaded: ${action.name} @ v${action.version}, ${fullFilePath}`, "debug");
                }
            };
            index_1.api.watchFileAndAct(fullFilePath, async () => {
                if (!index_1.api.config.general.developmentModeForceRestart) {
                    // reload by updating in-memory copy of our action
                    index_1.api.actions.loadFile(fullFilePath, true);
                    index_1.api.params.buildPostVariables();
                    index_1.api.routes.loadRoutes();
                }
                else {
                    index_1.api.log(`*** Rebooting due to action change (${fullFilePath}) ***`, "info");
                    await index_1.api.commands.restart();
                }
            });
            let action;
            try {
                let collection = require(fullFilePath);
                if (typeof collection === "function") {
                    collection = [collection];
                }
                for (const i in collection) {
                    action = new collection[i]();
                    await action.validate(index_1.api);
                    if (!index_1.api.actions.actions[action.name]) {
                        index_1.api.actions.actions[action.name] = {};
                    }
                    if (!index_1.api.actions.versions[action.name]) {
                        index_1.api.actions.versions[action.name] = [];
                    }
                    if (index_1.api.actions.actions[action.name][action.version] && !reload) {
                        index_1.api.log(`an existing action with the same name \`${action.name}\` will be overridden by the file ${fullFilePath}`, "warning");
                    }
                    index_1.api.actions.actions[action.name][action.version] = action;
                    index_1.api.actions.versions[action.name].push(action.version);
                    index_1.api.actions.versions[action.name].sort();
                    loadMessage(action);
                }
            }
            catch (error) {
                try {
                    index_1.api.exceptionHandlers.loader(fullFilePath, error);
                    delete index_1.api.actions.actions[action.name][action.version];
                }
                catch (_error) {
                    throw error;
                }
            }
        };
        for (const i in index_1.api.config.general.paths.action) {
            const p = index_1.api.config.general.paths.action[i];
            let files = glob.sync(path.join(p, "**", "**/*(*.js|*.ts)"));
            files = index_1.api.utils.ensureNoTsHeaderFiles(files);
            for (const j in files) {
                await index_1.api.actions.loadFile(files[j]);
            }
        }
        for (const pluginName in index_1.api.config.plugins) {
            if (index_1.api.config.plugins[pluginName].actions !== false) {
                const pluginPath = index_1.api.config.plugins[pluginName].path;
                let files = glob.sync(path.join(pluginPath, "actions", "**", "**/*(*.js|*.ts)"));
                files = index_1.api.utils.ensureNoTsHeaderFiles(files);
                for (const j in files) {
                    await index_1.api.actions.loadFile(files[j]);
                }
            }
        }
    }
}
exports.Actions = Actions;
