"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const index_1 = require("../index");
/**
 * Chat & Realtime Communication Methods
 */
class ChatRoom extends index_1.Initializer {
    constructor() {
        super();
        this.name = "chatRoom";
        this.loadPriority = 520;
        this.startPriority = 200;
    }
    async initialize() {
        if (index_1.api.config.redis.enabled === false) {
            return;
        }
        index_1.api.chatRoom = {
            middleware: {},
            globalMiddleware: [],
            messageChannel: "/actionhero/chat/chat",
            keys: {
                rooms: "actionhero:chatRoom:rooms",
                members: "actionhero:chatRoom:members:"
            }
        };
        /**
         * Add a middleware component to connection handling.
         */
        index_1.api.chatRoom.addMiddleware = (data) => {
            if (!data.name) {
                throw new Error("middleware.name is required");
            }
            if (!data.priority) {
                data.priority = index_1.api.config.general.defaultMiddlewarePriority;
            }
            data.priority = Number(data.priority);
            index_1.api.chatRoom.middleware[data.name] = data;
            index_1.api.chatRoom.globalMiddleware.push(data.name);
            index_1.api.chatRoom.globalMiddleware.sort((a, b) => {
                if (index_1.api.chatRoom.middleware[a].priority >
                    index_1.api.chatRoom.middleware[b].priority) {
                    return 1;
                }
                else {
                    return -1;
                }
            });
        };
        /**
         * Send a message to all members of a chat room.  This is used by the server.
         */
        index_1.api.chatRoom.broadcast = async (connection, room, message) => {
            if (!room || !message) {
                throw new Error(index_1.api.config.errors.connectionRoomAndMessage(connection));
            }
            else if (connection.rooms === undefined ||
                connection.rooms.indexOf(room) > -1) {
                const payload = {
                    messageType: "chat",
                    serverToken: index_1.api.config.general.serverToken,
                    serverId: index_1.api.id,
                    message: message,
                    sentAt: new Date().getTime(),
                    connection: {
                        id: connection.id || "0",
                        room: room
                    }
                };
                const messagePayload = index_1.api.chatRoom.generateMessagePayload(payload);
                const newPayload = await index_1.api.chatRoom.runMiddleware(connection, messagePayload.room, "onSayReceive", messagePayload);
                const payloadToSend = {
                    messageType: "chat",
                    serverToken: index_1.api.config.general.serverToken,
                    serverId: index_1.api.id,
                    message: newPayload.message,
                    sentAt: newPayload.sentAt,
                    connection: {
                        id: newPayload.from,
                        room: newPayload.room
                    }
                };
                await index_1.api.redis.publish(payloadToSend);
            }
            else {
                throw new Error(index_1.api.config.errors.connectionNotInRoom(connection, room));
            }
        };
        index_1.api.chatRoom.generateMessagePayload = message => {
            return {
                message: message.message,
                room: message.connection.room,
                from: message.connection.id,
                context: "user",
                sentAt: message.sentAt
            };
        };
        index_1.api.chatRoom.incomingMessage = message => {
            const messagePayload = index_1.api.chatRoom.generateMessagePayload(message);
            Object.keys(index_1.api.connections.connections).forEach(connetionId => {
                const connection = index_1.api.connections.connections[connetionId];
                // we can parallize this, no need to await
                index_1.api.chatRoom.incomingMessagePerConnection(connection, messagePayload);
            });
        };
        index_1.api.chatRoom.incomingMessagePerConnection = async (connection, messagePayload) => {
            if (connection.canChat === true &&
                connection.rooms.indexOf(messagePayload.room) > -1) {
                try {
                    const newMessagePayload = await index_1.api.chatRoom.runMiddleware(connection, messagePayload.room, "say", messagePayload);
                    connection.sendMessage(newMessagePayload, "say");
                }
                catch (error) {
                    index_1.api.log(error, "warning", { messagePayload, connection });
                }
            }
        };
        /**
         * List all chat rooms created
         */
        index_1.api.chatRoom.list = async () => {
            return index_1.api.redis.clients.client.smembers(index_1.api.chatRoom.keys.rooms);
        };
        /**
         * Add a new chat room.  Throws an error if the room already exists.
         */
        index_1.api.chatRoom.add = async (room) => {
            const found = await index_1.api.chatRoom.exists(room);
            if (found === false) {
                return index_1.api.redis.clients.client.sadd(index_1.api.chatRoom.keys.rooms, room);
            }
            else {
                throw new Error(await index_1.api.config.errors.connectionRoomExists(room));
            }
        };
        /**
         * Remove an exsitng chat room.  All connections in the room will be removed.  Throws an error if the room does not exist.
         */
        index_1.api.chatRoom.destroy = async (room) => {
            const found = await index_1.api.chatRoom.exists(room);
            if (found === true) {
                await index_1.api.chatRoom.broadcast({}, room, await index_1.api.config.errors.connectionRoomHasBeenDeleted(room));
                const membersHash = await index_1.api.redis.clients.client.hgetall(index_1.api.chatRoom.keys.members + room);
                for (const id in membersHash) {
                    await index_1.api.chatRoom.removeMember(id, room, false);
                }
                await index_1.api.redis.clients.client.srem(index_1.api.chatRoom.keys.rooms, room);
                await index_1.api.redis.clients.client.del(index_1.api.chatRoom.keys.members + room);
            }
            else {
                throw new Error(await index_1.api.config.errors.connectionRoomNotExist(room));
            }
        };
        /**
         * Check if a room exists.
         */
        index_1.api.chatRoom.exists = async (room) => {
            const bool = await index_1.api.redis.clients.client.sismember(index_1.api.chatRoom.keys.rooms, room);
            let found = false;
            if (bool === 1 || bool === true) {
                found = true;
            }
            return found;
        };
        /**
         * An overwritable method which configures what properties of connections in a room to return via `api.chatRoom.roomStatus`
         */
        index_1.api.chatRoom.sanitizeMemberDetails = memberData => {
            return {
                id: memberData.id,
                joinedAt: memberData.joinedAt
            };
        };
        /**
         * Learn about the connections in the room.
         * Returns a hash of the form { room: room, members: cleanedMembers, membersCount: count }.  Members is an array of connections in the room sanitized via `api.chatRoom.sanitizeMemberDetails`
         */
        index_1.api.chatRoom.roomStatus = async (room) => {
            if (room) {
                const found = await index_1.api.chatRoom.exists(room);
                if (found === true) {
                    const key = index_1.api.chatRoom.keys.members + room;
                    const members = await index_1.api.redis.clients.client.hgetall(key);
                    const cleanedMembers = {};
                    let count = 0;
                    for (const id in members) {
                        const data = JSON.parse(members[id]);
                        cleanedMembers[id] = index_1.api.chatRoom.sanitizeMemberDetails(data);
                        count++;
                    }
                    return {
                        room: room,
                        members: cleanedMembers,
                        membersCount: count
                    };
                }
                else {
                    throw new Error(await index_1.api.config.errors.connectionRoomNotExist(room));
                }
            }
            else {
                throw new Error(await index_1.api.config.errors.connectionRoomRequired());
            }
        };
        /**
         * An overwritable method which configures what properties of connections in a room are initially stored about a connection when added via `api.chatRoom.addMember`
         */
        index_1.api.chatRoom.generateMemberDetails = (connection) => {
            return {
                id: connection.id,
                joinedAt: new Date().getTime(),
                host: index_1.api.id
            };
        };
        /**
         * Add a connection (via id) to a rooom.  Throws errors if the room does not exist, or the connection is already in the room.  Middleware errors also throw.
         */
        index_1.api.chatRoom.addMember = async (connectionId, room) => {
            const connection = index_1.api.connections.connections[connectionId];
            if (!connection) {
                return index_1.api.redis.doCluster("api.chatRoom.addMember", [connectionId, room], connectionId, true);
            }
            if (connection.rooms.indexOf(room) >= 0) {
                throw new Error(await index_1.api.config.errors.connectionAlreadyInRoom(connection, room));
            }
            if (connection.rooms.indexOf(room) < 0) {
                const found = await index_1.api.chatRoom.exists(room);
                if (!found) {
                    throw new Error(await index_1.api.config.errors.connectionRoomNotExist(room));
                }
            }
            if (connection.rooms.indexOf(room) < 0) {
                await index_1.api.chatRoom.runMiddleware(connection, room, "join");
            }
            if (connection.rooms.indexOf(room) < 0) {
                const memberDetails = index_1.api.chatRoom.generateMemberDetails(connection);
                connection.rooms.push(room);
                await index_1.api.redis.clients.client.hset(index_1.api.chatRoom.keys.members + room, connection.id, JSON.stringify(memberDetails));
            }
            return true;
        };
        /**
         * Remote a connection (via id) from a rooom.  Throws errors if the room does not exist, or the connection is not in the room.  Middleware errors also throw.
         * toWaitRemote: Should this method wait until the remote ActionHero server (the one the connection is connected too) responds?
         */
        index_1.api.chatRoom.removeMember = async (connectionId, room, toWaitRemote = true) => {
            const connection = index_1.api.connections.connections[connectionId];
            if (!connection) {
                return index_1.api.redis.doCluster("api.chatRoom.removeMember", [connectionId, room], connectionId, toWaitRemote);
            }
            if (connection.rooms.indexOf(room) < 0) {
                throw new Error(await index_1.api.config.errors.connectionNotInRoom(connection, room));
            }
            if (connection.rooms.indexOf(room) >= 0) {
                const found = await index_1.api.chatRoom.exists(room);
                if (!found) {
                    throw new Error(await index_1.api.config.errors.connectionRoomNotExist(room));
                }
            }
            if (connection.rooms.indexOf(room) >= 0) {
                await index_1.api.chatRoom.runMiddleware(connection, room, "leave");
            }
            if (connection.rooms.indexOf(room) >= 0) {
                const index = connection.rooms.indexOf(room);
                connection.rooms.splice(index, 1);
                await index_1.api.redis.clients.client.hdel(index_1.api.chatRoom.keys.members + room, connection.id);
            }
            return true;
        };
        index_1.api.chatRoom.runMiddleware = async (connection, room, direction, messagePayload) => {
            let newMessagePayload;
            if (messagePayload) {
                newMessagePayload = Object.assign({}, messagePayload);
            }
            for (const name of index_1.api.chatRoom.globalMiddleware) {
                const m = index_1.api.chatRoom.middleware[name];
                if (typeof m[direction] === "function") {
                    if (messagePayload) {
                        const data = await m[direction](connection, room, newMessagePayload);
                        if (data) {
                            newMessagePayload = data;
                        }
                    }
                    else {
                        await m[direction](connection, room);
                    }
                }
            }
            return newMessagePayload;
        };
    }
    async start() {
        if (index_1.api.config.redis.enabled === false) {
            return;
        }
        index_1.api.redis.subscriptionHandlers.chat = message => {
            if (index_1.api.chatRoom) {
                index_1.api.chatRoom.incomingMessage(message);
            }
        };
        if (index_1.api.config.general.startingChatRooms) {
            const rooms = Object.keys(index_1.api.config.general.startingChatRooms);
            for (const room of rooms) {
                index_1.api.log(`ensuring the existence of the chatRoom: ${room}`);
                try {
                    await index_1.api.chatRoom.add(room);
                }
                catch (error) {
                    if (!error
                        .toString()
                        .match(await index_1.api.config.errors.connectionRoomExists(room))) {
                        throw error;
                    }
                }
            }
        }
    }
}
exports.ChatRoom = ChatRoom;
