"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const uuid = require("uuid");
const index_1 = require("../index");
/**
 * Redis helpers and connections.
 */
class Redis extends index_1.Initializer {
    constructor() {
        super();
        this.name = "redis";
        this.loadPriority = 200;
        this.startPriority = 101;
        this.stopPriority = 99999;
    }
    async initialize() {
        if (index_1.api.config.redis.enabled === false) {
            return;
        }
        index_1.api.redis = {};
        index_1.api.redis.clients = {};
        index_1.api.redis.subscriptionHandlers = {};
        index_1.api.redis.rpcCallbacks = {};
        index_1.api.redis.status = {
            subscribed: false
        };
        /**
         * Publish a message to all other ActionHero nodes in the clsuter.  Will be autneticated against `api.config.serverToken`
         * ```js
         * let payload = {
         *   messageType: 'myMessageType',
         *   serverId: api.id,
         *   serverToken: api.config.general.serverToken,
         *   message: 'hello!'
         * }
         * await api.redis.publish(payload)
         * ```
         */
        index_1.api.redis.publish = async (payload) => {
            const channel = index_1.api.config.general.channel;
            return index_1.api.redis.clients.client.publish(channel, JSON.stringify(payload));
        };
        index_1.api.redis.subscriptionHandlers.do = async (message) => {
            if (!message.connectionId ||
                (index_1.api.connections && index_1.api.connections.connections[message.connectionId])) {
                const cmdParts = message.method.split(".");
                const cmd = cmdParts.shift();
                if (cmd !== "api") {
                    throw new Error("cannot operate on a method outside of the api object");
                }
                const method = index_1.api.utils.dotProp.get(index_1.api, cmdParts.join("."));
                let args = message.args;
                if (args === null) {
                    args = [];
                }
                if (!Array.isArray(args)) {
                    args = [args];
                }
                if (method) {
                    const response = await method.apply(null, args);
                    await index_1.api.redis.respondCluster(message.messageId, response);
                }
                else {
                    index_1.api.log("RPC method `" + cmdParts.join(".") + "` not found", "warning");
                }
            }
        };
        index_1.api.redis.subscriptionHandlers.doResponse = function (message) {
            if (index_1.api.redis.rpcCallbacks[message.messageId]) {
                const { resolve, timer } = index_1.api.redis.rpcCallbacks[message.messageId];
                clearTimeout(timer);
                resolve(message.response);
                delete index_1.api.redis.rpcCallbacks[message.messageId];
            }
        };
        /**
         * Invoke a command on all servers in this cluster.
         */
        index_1.api.redis.doCluster = async (method, args = [], connectionId, waitForResponse = false) => {
            const messageId = uuid.v4();
            const payload = {
                messageType: "do",
                serverId: index_1.api.id,
                serverToken: index_1.api.config.general.serverToken,
                messageId: messageId,
                method: method,
                connectionId: connectionId,
                args: args // [1,2,3]
            };
            // we need to be sure that we build the response-handling promise before sending the request to Redis
            // it is possible for another node to get and work the request before we resolve our write
            // see https://github.com/actionhero/actionhero/issues/1244 for more information
            if (waitForResponse) {
                return new Promise(async (resolve, reject) => {
                    const timer = setTimeout(() => reject(new Error("RPC Timeout")), index_1.api.config.general.rpcTimeout);
                    index_1.api.redis.rpcCallbacks[messageId] = { timer, resolve, reject };
                    try {
                        await index_1.api.redis.publish(payload);
                    }
                    catch (e) {
                        clearTimeout(timer);
                        delete index_1.api.redis.rpcCallbacks[messageId];
                        throw e;
                    }
                });
            }
            await index_1.api.redis.publish(payload);
        };
        index_1.api.redis.respondCluster = async (messageId, response) => {
            const payload = {
                messageType: "doResponse",
                serverId: index_1.api.id,
                serverToken: index_1.api.config.general.serverToken,
                messageId: messageId,
                response: response // args to pass back, including error
            };
            await index_1.api.redis.publish(payload);
        };
        const connectionNames = ["client", "subscriber", "tasks"];
        for (var i in connectionNames) {
            const r = connectionNames[i];
            if (index_1.api.config.redis[r].buildNew === true) {
                const args = index_1.api.config.redis[r].args;
                index_1.api.redis.clients[r] = new index_1.api.config.redis[r].konstructor(args[0], args[1], args[2]);
                index_1.api.redis.clients[r].on("error", error => {
                    index_1.api.log(`Redis connection \`${r}\` error`, "alert", error);
                });
                index_1.api.redis.clients[r].on("connect", () => {
                    index_1.api.log(`Redis connection \`${r}\` connected`, "debug");
                });
            }
            else {
                index_1.api.redis.clients[r] = index_1.api.config.redis[r].konstructor.apply(null, index_1.api.config.redis[r].args);
                index_1.api.redis.clients[r].on("error", error => {
                    index_1.api.log(`Redis connection \`${r}\` error`, "alert", error);
                });
                index_1.api.log(`Redis connection \`${r}\` connected`, "debug");
            }
            await index_1.api.redis.clients[r].get("_test");
        }
        if (!index_1.api.redis.status.subscribed) {
            await index_1.api.redis.clients.subscriber.subscribe(index_1.api.config.general.channel);
            index_1.api.redis.status.subscribed = true;
            const messageHandler = async (messageChannel, stringifiedMessage) => {
                let message;
                try {
                    message = JSON.parse(stringifiedMessage);
                }
                catch (e) {
                    message = {};
                }
                if (messageChannel === index_1.api.config.general.channel &&
                    message.serverToken === index_1.api.config.general.serverToken) {
                    if (index_1.api.redis.subscriptionHandlers[message.messageType]) {
                        await index_1.api.redis.subscriptionHandlers[message.messageType](message);
                    }
                }
            };
            index_1.api.redis.clients.subscriber.on("message", messageHandler);
        }
    }
    async start() {
        if (index_1.api.config.redis.enabled === false) {
            index_1.api.log("redis is disabled", "notice");
        }
        else {
            await index_1.api.redis.doCluster("api.log", [
                `actionhero member ${index_1.api.id} has joined the cluster`
            ]);
        }
    }
    async stop() {
        if (index_1.api.config.redis.enabled === false) {
            return;
        }
        await index_1.api.redis.clients.subscriber.unsubscribe();
        index_1.api.redis.status.subscribed = false;
        await index_1.api.redis.doCluster("api.log", [
            `actionhero member ${index_1.api.id} has left the cluster`
        ]);
        const keys = Object.keys(index_1.api.redis.clients);
        for (const i in keys) {
            const client = index_1.api.redis.clients[keys[i]];
            if (typeof client.quit === "function") {
                await client.quit();
            }
            else if (typeof client.end === "function") {
                await client.end();
            }
            else if (typeof client.disconnect === "function") {
                await client.disconnect();
            }
        }
    }
}
exports.Redis = Redis;
