"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const glob = require("glob");
const index_1 = require("../index");
/**
 * Manages the servers in this ActionHero instance.
 */
class Servers extends index_1.Initializer {
    constructor() {
        super();
        this.name = "servers";
        this.loadPriority = 599;
        this.startPriority = 900;
        this.stopPriority = 100;
    }
    async initialize() {
        index_1.api.servers = {
            servers: {}
        };
        const serverFolders = [path.resolve(path.join(__dirname, "..", "servers"))];
        index_1.api.config.general.paths.server.forEach(p => {
            p = path.resolve(p);
            if (serverFolders.indexOf(p) < 0) {
                serverFolders.push(p);
            }
        });
        for (const i in serverFolders) {
            const p = serverFolders[i];
            let files = glob.sync(path.join(p, "**", "**/*(*.js|*.ts)"));
            for (const pluginName in index_1.api.config.plugins) {
                if (index_1.api.config.plugins[pluginName].servers !== false) {
                    const pluginPath = index_1.api.config.plugins[pluginName].path;
                    files = files.concat(glob.sync(path.join(pluginPath, "servers", "**", "**/*(*.js|*.ts)")));
                }
            }
            files = index_1.api.utils.ensureNoTsHeaderFiles(files);
            for (const j in files) {
                const filename = files[j];
                const ExportedClasses = require(filename);
                if (Object.keys(ExportedClasses).length > 1) {
                    throw new Error(`server file ${filename} exports more than one server`);
                }
                const server = new ExportedClasses[Object.keys(ExportedClasses)[0]]();
                server.config = index_1.api.config.servers[server.type]; // shorthand access
                if (server.config && server.config.enabled === true) {
                    await server.initialize();
                    if (index_1.api.servers.servers[server.type]) {
                        index_1.api.log(`an existing server with the same type \`${server.type}\` will be overridden by the file ${filename}`, "warning");
                    }
                    index_1.api.servers.servers[server.type] = server;
                    index_1.api.log(`Initialized server: ${server.type}`, "debug");
                }
                index_1.api.watchFileAndAct(filename, () => {
                    index_1.api.log(`*** Rebooting due to server (${server.type}) change ***`, "info");
                    index_1.api.commands.restart();
                });
            }
        }
    }
    async start() {
        const serverNames = Object.keys(index_1.api.servers.servers);
        for (const i in serverNames) {
            const serverName = serverNames[i];
            const server = index_1.api.servers.servers[serverName];
            if (server && server.config.enabled === true) {
                let message = "";
                message += `Starting server: \`${serverName}\``;
                if (index_1.api.config.servers[serverName].bindIP) {
                    message += ` @ ${index_1.api.config.servers[serverName].bindIP}`;
                }
                if (index_1.api.config.servers[serverName].port) {
                    message += `:${index_1.api.config.servers[serverName].port}`;
                }
                index_1.api.log(message, "notice");
                await server.start();
                index_1.api.log(`Server started: ${serverName}`, "debug");
            }
        }
    }
    async stop() {
        const serverNames = Object.keys(index_1.api.servers.servers);
        for (const i in serverNames) {
            const serverName = serverNames[i];
            const server = index_1.api.servers.servers[serverName];
            if ((server && server.config.enabled === true) || !server) {
                index_1.api.log(`Stopping server: ${serverName}`, "notice");
                await server.stop();
                server.removeAllListeners();
                index_1.api.log(`Server stopped: ${serverName}`, "debug");
            }
        }
    }
}
exports.Servers = Servers;
