"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const glob = require("glob");
const path = require("path");
const node_resque_1 = require("node-resque");
const index_1 = require("../index");
/**
 * Tools for enquing and inspecting the task sytem (delayed jobs).
 */
class Tasks extends index_1.Initializer {
    constructor() {
        super();
        this.name = "tasks";
        this.loadPriority = 699;
        this.startPriority = 900;
    }
    async initialize() {
        index_1.api.tasks = {
            tasks: {},
            jobs: {},
            middleware: {},
            globalMiddleware: []
        };
        index_1.api.tasks.loadFile = (fullFilePath, reload = false) => {
            index_1.api.watchFileAndAct(fullFilePath, async () => {
                if (!index_1.api.config.general.developmentModeForceRestart) {
                    // reload by updating in-memory copy of our task
                    index_1.api.tasks.loadFile(fullFilePath, true);
                }
                else {
                    index_1.api.log(`*** Rebooting due to task change (${fullFilePath}) ***`, "info");
                    await index_1.api.commands.restart();
                }
            });
            let task;
            let collection = require(fullFilePath);
            for (const i in collection) {
                const TaskClass = collection[i];
                task = new TaskClass();
                task.validate();
                if (index_1.api.tasks.tasks[task.name] && !reload) {
                    index_1.api.log(`an existing task with the same name \`${task.name}\` will be overridden by the file ${fullFilePath}`, "warning");
                }
                index_1.api.tasks.tasks[task.name] = task;
                index_1.api.tasks.jobs[task.name] = index_1.api.tasks.jobWrapper(task.name);
                index_1.api.log(`task ${reload ? "(re)" : ""} loaded: ${task.name}, ${fullFilePath}`, reload ? "info" : "debug");
            }
        };
        /**
         * @private
         */
        index_1.api.tasks.jobWrapper = (taskName) => {
            const task = index_1.api.tasks.tasks[taskName];
            const middleware = task.middleware || [];
            const plugins = task.plugins || [];
            const pluginOptions = task.pluginOptions || [];
            if (task.frequency > 0) {
                if (plugins.indexOf("JobLock") < 0) {
                    plugins.push("JobLock");
                }
                if (plugins.indexOf("QueueLock") < 0) {
                    plugins.push("QueueLock");
                }
                if (plugins.indexOf("DelayQueueLock") < 0) {
                    plugins.push("DelayQueueLock");
                }
            }
            // load middleware into plugins
            const processMiddleware = m => {
                if (index_1.api.tasks.middleware[m]) {
                    //@ts-ignore
                    class NodeResquePlugin extends node_resque_1.Plugin {
                        constructor(...args) {
                            //@ts-ignore
                            super(...args);
                            if (index_1.api.tasks.middleware[m].preProcessor) {
                                this.beforePerform = index_1.api.tasks.middleware[m].preProcessor;
                            }
                            if (index_1.api.tasks.middleware[m].postProcessor) {
                                this.afterPerform = index_1.api.tasks.middleware[m].postProcessor;
                            }
                            if (index_1.api.tasks.middleware[m].preEnqueue) {
                                this.beforeEnqueue = index_1.api.tasks.middleware[m].preEnqueue;
                            }
                            if (index_1.api.tasks.middleware[m].postEnqueue) {
                                this.afterEnqueue = index_1.api.tasks.middleware[m].postEnqueue;
                            }
                        }
                    }
                    plugins.push(NodeResquePlugin);
                }
            };
            index_1.api.tasks.globalMiddleware.forEach(processMiddleware);
            middleware.forEach(processMiddleware);
            return {
                plugins: plugins,
                pluginOptions: pluginOptions,
                perform: async function () {
                    const combinedArgs = [].concat(Array.prototype.slice.call(arguments));
                    combinedArgs.push(this);
                    let response = null;
                    try {
                        response = await task.run.apply(task, combinedArgs);
                        await index_1.api.tasks.enqueueRecurrentTask(taskName);
                    }
                    catch (error) {
                        if (task.frequency > 0 && task.reEnqueuePeriodicTaskIfException) {
                            await index_1.api.tasks.enqueueRecurrentTask(taskName);
                        }
                        throw error;
                    }
                    return response;
                }
            };
        };
        /**
         * Enqueue a task to be performed in the background.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.enqueue = async (taskName, params = {}, queue = index_1.api.tasks.tasks[taskName].queue) => {
            return index_1.api.resque.queue.enqueue(queue, taskName, params);
        };
        /**
         * Enqueue a task to be performed in the background, at a certain time in the future.
         * Will throw an error if redis cannot be reached.
         *
         * Inputs:
         * * taskName: The name of the task.
         * * params: Params to pass to the task.
         * * queue: (Optional) Which queue/priority to run this instance of the task on.
         */
        index_1.api.tasks.enqueueAt = async (timestamp, taskName, params = {}, queue = index_1.api.tasks.tasks[taskName].queue) => {
            return index_1.api.resque.queue.enqueueAt(timestamp, queue, taskName, params);
        };
        /**
         * Enqueue a task to be performed in the background, at a certain number of ms from now.
         * Will throw an error if redis cannot be reached.
         *
         * Inputs:
         * * timestamp: At what time the task is able to be run.  Does not gaurentee that the task will be run at this time. (in ms)
         * * taskName: The name of the task.
         * * params: Params to pass to the task.
         * * queue: (Optional) Which queue/priority to run this instance of the task on.
         */
        index_1.api.tasks.enqueueIn = async (time, taskName, params = {}, queue = index_1.api.tasks.tasks[taskName].queue) => {
            return index_1.api.resque.queue.enqueueIn(time, queue, taskName, params);
        };
        /**
         * Delete a previously enqueued task, which hasn't been run yet, from a queue.
         * Will throw an error if redis cannot be reached.
         *
         * Inputs:
         * * q: Which queue/priority is the task stored on?
         * * taskName: The name of the job, likley to be the same name as a tak.
         * * args: The arguments of the job.  Note, arguments passed to a Task initially may be modified when enqueuing.  It is best to read job properties first via `api.tasks.queued` or similar method.
         * * count: Of the jobs that match q, taskName, and args, up to what position should we delete? (Default 0; this command is 0-indexed)
         */
        index_1.api.tasks.del = async (q, taskName, args, count) => {
            return index_1.api.resque.queue.del(q, taskName, args, count);
        };
        /**
         * Delete all previously enqueued tasks, which haven't been run yet, from all possible delayed timestamps.
         * Will throw an error if redis cannot be reached.
         *
         * Inputs:
         * * q: Which queue/priority is to run on?
         * * taskName: The name of the job, likley to be the same name as a tak.
         * * args  The arguments of the job.  Note, arguments passed to a Task initially may be modified when enqueuing. It is best to read job properties first via `api.tasks.delayedAt` or similar method.
         */
        index_1.api.tasks.delDelayed = async (q, taskName, args) => {
            return index_1.api.resque.queue.delDelayed(q, taskName, args);
        };
        /**
         * Return the timestamps a task is scheduled for.
         * Will throw an error if redis cannot be reached.
         *
         * Inputs:
         * * q: Which queue/priority is to run on?
         * * taskName: The name of the job, likley to be the same name as a tak.
         * * args: The arguments of the job.  Note, arguments passed to a Task initially may be modified when enqueuing.  It is best to read job properties first via `api.tasks.delayedAt` or similar method.
         * @return {Promise<Array>}    Returns an array of timestamps.
         */
        index_1.api.tasks.scheduledAt = async (q, taskName, args) => {
            return index_1.api.resque.queue.scheduledAt(q, taskName, args);
        };
        /**
         * Return all resque stats for this namespace (how jobs failed, jobs succeded, etc)
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.stats = async () => {
            return index_1.api.resque.queue.stats();
        };
        /**
         * Retrieve the details of jobs enqueued on a certain queue between start and stop (0-indexed)
         * Will throw an error if redis cannot be reached.
         *
         * Inputs:
         * * q      The name of the queue.
         * * start  The index of the first job to return.
         * * stop   The index of the last job to return.
         */
        index_1.api.tasks.queued = async (q, start, stop) => {
            return index_1.api.resque.queue.queued(q, start, stop);
        };
        /**
         * Delete a queue in redis, and all jobs stored on it.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.delQueue = async (q) => {
            return index_1.api.resque.queue.delQueue(q);
        };
        /**
         * Return any locks, as created by resque plugins or task middleware, in this redis namespace.
         * Will contain locks with keys like `resque:lock:{job}` and `resque:workerslock:{workerId}`
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.locks = async () => {
            return index_1.api.resque.queue.locks();
        };
        /**
         * Delete a lock on a job or worker.  Locks can be found via `api.tasks.locks`
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.delLock = async (lock) => {
            return index_1.api.resque.queue.delLock(lock);
        };
        /**
         * List all timestamps for which tasks are enqueued in the future, via `api.tasks.enqueueIn` or `api.tasks.enqueueAt`
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.timestamps = async () => {
            return index_1.api.resque.queue.timestamps();
        };
        /**
         * Return all jobs which have been enqueued to run at a certain timestamp.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.delayedAt = async (timestamp) => {
            return index_1.api.resque.queue.delayedAt(timestamp);
        };
        /**
         * Retrun all delayed jobs, orginized by the timetsamp at where they are to run at.
         * Note: This is a very slow command.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.allDelayed = async () => {
            return index_1.api.resque.queue.allDelayed();
        };
        /**
         * Retrun all workers registered by all members of this cluster.
         * Note: MultiWorker processors each register as a unique worker.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.workers = async () => {
            return index_1.api.resque.queue.workers();
        };
        /**
         * What is a given worker working on?  If the worker is idle, 'started' will be returned.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.workingOn = async (workerName, queues) => {
            return index_1.api.resque.queue.workingOn(workerName, queues);
        };
        /**
         * Return all workers and what job they might be working on.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.allWorkingOn = async () => {
            return index_1.api.resque.queue.allWorkingOn();
        };
        /**
         * How many jobs are in the failed queue.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.failedCount = async () => {
            return index_1.api.resque.queue.failedCount();
        };
        /**
         * Retrieve the details of failed jobs between start and stop (0-indexed).
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.failed = async (start, stop) => {
            return index_1.api.resque.queue.failed(start, stop);
        };
        /**
         * Remove a specific job from the failed queue.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.removeFailed = async (failedJob) => {
            return index_1.api.resque.queue.removeFailed(failedJob);
        };
        /**
         * Remove a specific job from the failed queue, and retry it by placing it back into its original queue.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.retryAndRemoveFailed = async (failedJob) => {
            return index_1.api.resque.queue.retryAndRemoveFailed(failedJob);
        };
        /**
         * If a worker process crashes, it will leave its state in redis as "working".
         * You can remove workers from redis you know to be over, by specificing an age which would make them too old to exist.
         * This method will remove the data created by a 'stuck' worker and move the payload to the error queue.
         * However, it will not actually remove any processes which may be running.  A job *may* be running that you have removed.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.cleanOldWorkers = async (age) => {
            return index_1.api.resque.queue.cleanOldWorkers(age);
        };
        /**
         * Ensures that a task which has a frequency is either running, or already enqueued.
         * This is run automatically at boot for all tasks which have a frequency, via `api.tasks.enqueueAllRecurrentTasks`.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.enqueueRecurrentTask = async (taskName) => {
            const task = index_1.api.tasks.tasks[taskName];
            if (task.frequency > 0) {
                await index_1.api.tasks.del(task.queue, taskName);
                await index_1.api.tasks.delDelayed(task.queue, taskName);
                await index_1.api.tasks.enqueueIn(task.frequency, taskName);
                index_1.api.log(`re-enqueued recurrent job ${taskName}`, index_1.api.config.tasks.schedulerLogging.reEnqueue);
            }
        };
        /**
         * This is run automatically at boot for all tasks which have a frequency, calling `api.tasks.enqueueRecurrentTask`
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.enqueueAllRecurrentTasks = async () => {
            const jobs = [];
            const loadedTasks = [];
            Object.keys(index_1.api.tasks.tasks).forEach(taskName => {
                const task = index_1.api.tasks.tasks[taskName];
                if (task.frequency > 0) {
                    jobs.push(async () => {
                        const toRun = await index_1.api.tasks.enqueue(taskName);
                        if (toRun === true) {
                            index_1.api.log(`enqueuing periodic task: ${taskName}`, index_1.api.config.tasks.schedulerLogging.enqueue);
                            loadedTasks.push(taskName);
                        }
                    });
                }
            });
            await index_1.api.utils.asyncWaterfall(jobs);
            return loadedTasks;
        };
        /**
         * Stop a task with a frequency by removing it from all possible queues.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.stopRecurrentTask = async (taskName) => {
            // find the jobs in either the normal queue or delayed queues
            const task = index_1.api.tasks.tasks[taskName];
            if (task.frequency > 0) {
                let removedCount = 0;
                const count = await index_1.api.tasks.del(task.queue, task.name, {}, 1);
                removedCount = removedCount + count;
                const timestamps = await index_1.api.tasks.delDelayed(task.queue, task.name, {});
                removedCount = removedCount + timestamps.length;
                return removedCount;
            }
        };
        /**
         * Return wholistic details about the task system, including failures, queues, and workers.
         * Will throw an error if redis cannot be reached.
         */
        index_1.api.tasks.details = async () => {
            const details = { queues: {}, workers: {}, stats: null };
            details.workers = await index_1.api.tasks.allWorkingOn();
            details.stats = await index_1.api.tasks.stats();
            const queues = await index_1.api.resque.queue.queues();
            for (const i in queues) {
                const queue = queues[i];
                const length = await index_1.api.resque.queue.length(queue);
                details.queues[queue] = { length: length };
            }
            return details;
        };
        index_1.api.tasks.loadTasks = reload => {
            index_1.api.config.general.paths.task.forEach(p => {
                index_1.api.utils
                    .ensureNoTsHeaderFiles(glob.sync(path.join(p, "**", "**/*(*.js|*.ts)")))
                    .forEach(f => {
                    index_1.api.tasks.loadFile(f, reload);
                });
            });
            for (const pluginName in index_1.api.config.plugins) {
                if (index_1.api.config.plugins[pluginName].tasks !== false) {
                    const pluginPath = index_1.api.config.plugins[pluginName].path;
                    index_1.api.utils
                        .ensureNoTsHeaderFiles(glob.sync(path.join(pluginPath, "tasks", "**", "**/*(*.js|*.ts)")))
                        .forEach(f => {
                        index_1.api.tasks.loadFile(f, reload);
                    });
                }
            }
        };
        index_1.api.tasks.addMiddleware = (middleware) => {
            if (!middleware.name) {
                throw new Error("middleware.name is required");
            }
            if (!middleware.priority) {
                middleware.priority = index_1.api.config.general.defaultMiddlewarePriority;
            }
            middleware.priority = Number(middleware.priority);
            index_1.api.tasks.middleware[middleware.name] = middleware;
            if (middleware.global === true) {
                index_1.api.tasks.globalMiddleware.push(middleware.name);
                index_1.api.utils.sortGlobalMiddleware(index_1.api.tasks.globalMiddleware, index_1.api.tasks.middleware);
            }
            index_1.api.tasks.loadTasks(true);
        };
        index_1.api.tasks.loadTasks(false);
    }
    async start() {
        if (index_1.api.config.redis.enabled === false) {
            return;
        }
        if (index_1.api.config.tasks.scheduler === true) {
            await index_1.api.tasks.enqueueAllRecurrentTasks();
        }
    }
}
exports.Tasks = Tasks;
