"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
const dotProp = require("dot-prop");
const os = require("os");
const index_1 = require("../index");
/**
 * Utilites for any ActionHero project.
 */
class Utils extends index_1.Initializer {
    constructor() {
        super();
        this.name = "utils";
        this.loadPriority = 1;
    }
    async initialize() {
        if (!index_1.api.utils) {
            index_1.api.utils = {};
        }
        index_1.api.utils.dotProp = dotProp;
        /**
         * In series, run an array of `async` functions
         *
         * without arguments
         * ```js
         * let sleepyFunc = async () => {
         *   await new Promise((resolve) => { setTimeout(resolve, 100) })
         *   return (new Date()).getTime()
         * }
         * let jobs = [sleepyFunc, sleepyFunc, sleepyFunc]
         * let responses = await api.utils.asyncWaterfall(jobs)
         * // responses = [1506536188356, 1506536188456, 1506536188456]
         * ```
         *
         * with arguments
         * ```js
         * let sleepyFunc = async (response) => {
         *   await new Promise((resolve) => { setTimeout(resolve, 100) })
         *   return response
         * }
         * let jobs = [
         *   {method: sleepyFunc, args: ['a']},
         *   {method: sleepyFunc, args: ['b']},
         *   {method: sleepyFunc, args: ['c']}
         * ]
         * let responses = await api.utils.asyncWaterfall(jobs)
         * // responses = ['a', 'b', 'c']
         * ```
         */
        index_1.api.utils.asyncWaterfall = async (jobs) => {
            const results = [];
            while (jobs.length > 0) {
                const collection = jobs.shift();
                let job;
                let args;
                if (typeof collection === "function") {
                    job = collection;
                    args = [];
                }
                else {
                    job = collection.method;
                    args = collection.args;
                }
                const value = await job.apply(this, args);
                results.push(value);
            }
            if (results.length === 0) {
                return null;
            }
            if (results.length === 1) {
                return results[0];
            }
            return results;
        };
        /**
         * Recursivley merge 2 Objects together.  Will resolve functions if they are present, unless the parent Object has the propery `_toExpand = false`.
         * ActionHero uses this internally to construct and resolve the config.
         * Matching keys in B override A.
         */
        index_1.api.utils.hashMerge = (a, b, arg) => {
            const c = {};
            let i;
            let response;
            for (i in a) {
                if (index_1.api.utils.isPlainObject(a[i])) {
                    // can't be anded into above condition, or empty objects will overwrite and not merge
                    // also make sure empty objects are created
                    c[i] =
                        Object.keys(a[i]).length > 0
                            ? index_1.api.utils.hashMerge(c[i], a[i], arg)
                            : {};
                }
                else {
                    if (typeof a[i] === "function") {
                        response = a[i](arg);
                        if (index_1.api.utils.isPlainObject(response)) {
                            c[i] = index_1.api.utils.hashMerge(c[i], response, arg);
                        }
                        else {
                            c[i] = response;
                        }
                    }
                    else {
                        // don't create first term if it is undefined or null
                        if (a[i] === undefined || a[i] === null) {
                        }
                        else
                            c[i] = a[i];
                    }
                }
            }
            for (i in b) {
                if (index_1.api.utils.isPlainObject(b[i])) {
                    // can't be anded into above condition, or empty objects will overwrite and not merge
                    if (Object.keys(b[i]).length > 0)
                        c[i] = index_1.api.utils.hashMerge(c[i], b[i], arg);
                    // make sure empty objects are only created, when no key exists yet
                    else if (!(i in c))
                        c[i] = {};
                }
                else {
                    if (typeof b[i] === "function") {
                        response = b[i](arg);
                        if (index_1.api.utils.isPlainObject(response)) {
                            c[i] = index_1.api.utils.hashMerge(c[i], response, arg);
                        }
                        else {
                            c[i] = response;
                        }
                    }
                    else {
                        // ignore second term if it is undefined
                        if (b[i] === undefined) {
                        }
                        else if (b[i] == null && i in c)
                            // delete second term/key if value is null and it already exists
                            delete c[i];
                        // normal assignments for everything else
                        else
                            c[i] = b[i];
                    }
                }
            }
            return c;
        };
        index_1.api.utils.isPlainObject = o => {
            const safeTypes = [
                Boolean,
                Number,
                String,
                Function,
                Array,
                Date,
                RegExp,
                Buffer
            ];
            const safeInstances = ["boolean", "number", "string", "function"];
            const expandPreventMatchKey = "_toExpand"; // set `_toExpand = false` within an object if you don't want to expand it
            let i;
            if (!o) {
                return false;
            }
            if (o instanceof Object === false) {
                return false;
            }
            for (i in safeTypes) {
                if (o instanceof safeTypes[i]) {
                    return false;
                }
            }
            for (i in safeInstances) {
                if (typeof o === safeInstances[i]) {
                    return false;
                }
            }
            if (o[expandPreventMatchKey] === false) {
                return false;
            }
            return o.toString() === "[object Object]";
        };
        /**
         * Return only the unique values in an Array.
         */
        index_1.api.utils.arrayUniqueify = (arr) => {
            const a = [];
            for (let i = 0; i < arr.length; i++) {
                for (let j = i + 1; j < arr.length; j++) {
                    if (arr[i] === arr[j]) {
                        j = ++i;
                    }
                }
                a.push(arr[i]);
            }
            return a;
        };
        index_1.api.utils.sourceRelativeLinkPath = (linkfile, pluginPaths) => {
            const type = fs.readFileSync(linkfile).toString();
            const pathParts = linkfile.split(path.sep);
            const name = pathParts[pathParts.length - 1].split(".")[0];
            const pathsToTry = pluginPaths.slice(0);
            let pluginRoot;
            pathsToTry.forEach(pluginPath => {
                const pluginPathAttempt = path.normalize(pluginPath + path.sep + name);
                try {
                    const stats = fs.lstatSync(pluginPathAttempt);
                    if (!pluginRoot && (stats.isDirectory() || stats.isSymbolicLink())) {
                        pluginRoot = pluginPathAttempt;
                    }
                }
                catch (e) { }
            });
            if (!pluginRoot) {
                return false;
            }
            const pluginSection = path.normalize(pluginRoot + path.sep + type);
            return pluginSection;
        };
        /**
         * Collapsses an Object with numerical keys (like `arguments` in a function) to an Array
         */
        index_1.api.utils.collapseObjectToArray = (obj) => {
            try {
                const keys = Object.keys(obj);
                if (keys.length < 1) {
                    return false;
                }
                if (keys[0] !== "0") {
                    return false;
                }
                if (keys[keys.length - 1] !== String(keys.length - 1)) {
                    return false;
                }
                const arr = [];
                for (const i in keys) {
                    const key = keys[i];
                    if (String(parseInt(key)) !== key) {
                        return false;
                    }
                    else {
                        arr.push(obj[key]);
                    }
                }
                return arr;
            }
            catch (e) {
                return false;
            }
        };
        /**
         * Returns this server's external/public IP address
         */
        index_1.api.utils.getExternalIPAddress = () => {
            const ifaces = os.networkInterfaces();
            let ip = null;
            for (const dev in ifaces) {
                ifaces[dev].forEach(details => {
                    if (details.family === "IPv4" && details.address !== "127.0.0.1") {
                        ip = details.address;
                    }
                });
            }
            return ip;
        };
        /**
         * Return ip and port information if defined in the header
         */
        index_1.api.utils.parseHeadersForClientAddress = (headers) => {
            let ip = null;
            let port = null;
            if (headers["x-forwarded-for"]) {
                let parts;
                let forwardedIp = headers["x-forwarded-for"].split(",")[0];
                if (forwardedIp.indexOf(".") >= 0 ||
                    (forwardedIp.indexOf(".") < 0 && forwardedIp.indexOf(":") < 0)) {
                    // IPv4
                    forwardedIp = forwardedIp.replace("::ffff:", ""); // remove any IPv6 information, ie: '::ffff:127.0.0.1'
                    parts = forwardedIp.split(":");
                    if (parts[0]) {
                        ip = parts[0];
                    }
                    if (parts[1]) {
                        port = parts[1];
                    }
                }
                else {
                    // IPv6
                    parts = index_1.api.utils.parseIPv6URI(forwardedIp);
                    if (parts.host) {
                        ip = parts.host;
                    }
                    if (parts.port) {
                        port = parts.port;
                    }
                }
            }
            if (headers["x-forwarded-port"]) {
                port = headers["x-forwarded-port"];
            }
            if (headers["x-real-ip"]) {
                // https://distinctplace.com/2014/04/23/story-behind-x-forwarded-for-and-x-real-ip-headers/
                ip = headers["x-real-ip"];
            }
            return { ip, port };
        };
        /**
         * Transform the cookie headers of a node HTTP `req` Object into a hash.
         */
        index_1.api.utils.parseCookies = (req) => {
            const cookies = {};
            if (req.headers.cookie) {
                req.headers.cookie.split(";").forEach(cookie => {
                    const parts = cookie.split("=");
                    cookies[parts[0].trim()] = (parts[1] || "").trim();
                });
            }
            return cookies;
        };
        /**
         * Parse an IPv6 address, returning both host and port.
         * see https://github.com/actionhero/actionhero/issues/275
         */
        index_1.api.utils.parseIPv6URI = (addr) => {
            let host = "::1";
            let port = "80";
            const regexp = new RegExp(/\[([0-9a-f:]+(?:%.+)?)]:([0-9]{1,5})/);
            // if we have brackets parse them and find a port
            if (addr.indexOf("[") > -1 && addr.indexOf("]") > -1) {
                const res = regexp.exec(addr);
                if (res === null) {
                    throw new Error("failed to parse address");
                }
                host = res[1];
                port = res[2];
            }
            else {
                host = addr;
            }
            return { host: host, port: parseInt(port, 10) };
        };
        /**
         *swap out d.ts files for the JS versions when running with ts-node
         */
        index_1.api.utils.ensureNoTsHeaderFiles = (files) => {
            return files.filter(f => {
                if (f.match(/.*\.d\.ts$/)) {
                    return false;
                }
                else {
                    return true;
                }
            });
        };
        /**
         * Returns the averge delay between a tick of the node.js event loop, as measured for N calls of `process.nextTick`
         */
        index_1.api.utils.eventLoopDelay = async (itterations) => {
            const jobs = [];
            if (!itterations) {
                throw new Error("itterations is required");
            }
            const sleepyFunc = async () => {
                return new Promise(resolve => {
                    const start = process.hrtime();
                    process.nextTick(() => {
                        const delta = process.hrtime(start);
                        const ms = delta[0] * 1000 + delta[1] / 1000000;
                        resolve(ms);
                    });
                });
            };
            let i = 0;
            while (i < itterations) {
                jobs.push(sleepyFunc);
                i++;
            }
            const results = await index_1.api.utils.asyncWaterfall(jobs);
            let sum = 0;
            results.forEach(t => {
                sum += t;
            });
            const avg = Math.round((sum / results.length) * 10000) / 1000;
            return avg;
        };
        /**
         * Sorts an Array of Objects with a priority key
         */
        index_1.api.utils.sortGlobalMiddleware = (globalMiddlewareList, middleware) => {
            globalMiddlewareList.sort((a, b) => {
                if (middleware[a].priority > middleware[b].priority) {
                    return 1;
                }
                else {
                    return -1;
                }
            });
        };
        /**
         * Prepares acton params for logging.
         * Hides any sensitieve data as defined by `api.config.general.filteredParams`
         * Truncates long strings via `api.config.logger.maxLogStringLength`
         */
        index_1.api.utils.filterObjectForLogging = (params) => {
            const filteredParams = {};
            for (const i in params) {
                if (index_1.api.utils.isPlainObject(params[i])) {
                    filteredParams[i] = Object.assign({}, params[i]);
                }
                else if (typeof params[i] === "string") {
                    filteredParams[i] = params[i].substring(0, index_1.api.config.logger.maxLogStringLength);
                }
                else {
                    filteredParams[i] = params[i];
                }
            }
            index_1.api.config.general.filteredParams.forEach(configParam => {
                if (index_1.api.utils.dotProp.get(params, configParam) !== undefined) {
                    index_1.api.utils.dotProp.set(filteredParams, configParam, "[FILTERED]");
                }
            });
            return filteredParams;
        };
        /**
         Compare the first n elements of an array with another, logner array
         */
        index_1.api.utils.arrayStartingMatch = (a, b) => {
            if (a.length === 0) {
                return false;
            }
            if (b.length === 0) {
                return false;
            }
            let matching = true;
            let i = 0;
            while (i < a.length) {
                if (a[i] !== b[i]) {
                    matching = false;
                }
                i++;
            }
            return matching;
        };
        /**
        Sleep with a Promise
        */
        index_1.api.utils.sleep = (time) => {
            return new Promise(resolve => {
                setTimeout(resolve, time);
            });
        };
        /**
         * Check if a directory exists.
         */
        index_1.api.utils.dirExists = (dir) => {
            try {
                const stats = fs.lstatSync(dir);
                return stats.isDirectory() || stats.isSymbolicLink();
            }
            catch (e) {
                return false;
            }
        };
        /**
         * Check if a file exists.
         */
        index_1.api.utils.fileExists = (file) => {
            try {
                const stats = fs.lstatSync(file);
                return stats.isFile() || stats.isSymbolicLink();
            }
            catch (e) {
                return false;
            }
        };
        /**
         * Create a directory, only if it doesn't exist yet.
         * Throws an error if the directory already exists, or encounters a filesystem problem.
         */
        index_1.api.utils.createDirSafely = (dir) => {
            if (index_1.api.utils.dirExists(dir)) {
                const error = new Error(`directory '${path.normalize(dir)}' already exists`);
                // @ts-ignore
                error.code = "EEXIST";
                throw error;
            }
            else {
                fs.mkdirSync(path.normalize(dir), "0766");
                return `created directory '${path.normalize(dir)}'`;
            }
        };
        /**
         * Create a file, only if it doesn't exist yet.
         * Throws an error if the file already exists, or encounters a filesystem problem.
         */
        index_1.api.utils.createFileSafely = (file, data, overwrite = false) => {
            if (index_1.api.utils.fileExists(file) && !overwrite) {
                const error = new Error(`file '${path.normalize(file)}' already exists`);
                // @ts-ignore
                error.code = "EEXIST";
                throw error;
            }
            else {
                let message = `wrote file '${path.normalize(file)}'`;
                if (overwrite && index_1.api.utils.fileExists(file)) {
                    message = ` - overwritten file '${path.normalize(file)}'`;
                }
                fs.writeFileSync(path.normalize(file), data);
                return message;
            }
        };
        /**
         * Create an ActionHero LinkFile, only if it doesn't exist yet.
         * Throws an error if the file already exists, or encounters a filesystem problem.
         */
        index_1.api.utils.createLinkfileSafely = (filePath, type) => {
            if (index_1.api.utils.fileExists(filePath)) {
                const error = new Error(`link file '${filePath}' already exists`);
                // @ts-ignore
                error.code = "EEXIST";
                throw error;
            }
            else {
                fs.writeFileSync(filePath, type);
                return `creating linkfile '${filePath}'`;
            }
        };
        /**
         * Remove an ActionHero LinkFile, only if it exists.
         * Throws an error if the file does not exist, or encounters a filesystem problem.
         */
        index_1.api.utils.removeLinkfileSafely = (filePath) => {
            if (!index_1.api.utils.fileExists(filePath)) {
                const error = new Error(`link file '${filePath}' doesn't exist`);
                // @ts-ignore
                error.code = "ENOEXIST";
                throw error;
            }
            else {
                fs.unlinkSync(filePath);
                return `removing linkfile '${filePath}'`;
            }
        };
        /**
         * Create a system symbolic link.
         * Throws an error if it encounters a filesystem problem.
         */
        index_1.api.utils.createSymlinkSafely = (destination, source) => {
            if (index_1.api.utils.dirExists(destination)) {
                const error = new Error(`symbolic link '${destination}' already exists`);
                // @ts-ignore
                error.code = "EEXIST";
                throw error;
            }
            else {
                fs.symlinkSync(source, destination, "dir");
                return `creating symbolic link '${destination}' => '${source}'`;
            }
        };
        /**
         * Used by generator functions running from your `dist`, it replacaes the path with your `src`
         * Relies on api.config.general.paths
         */
        index_1.api.utils.replaceDistWithSrc = (f) => {
            return f.replace(index_1.api.config.general.paths.dist, index_1.api.config.general.paths.src);
        };
    }
}
exports.Utils = Utils;
